@@ -1,25 +1,32 @@
 //! Box selection for transaction inputs
 
 mod simple;
+
 use std::collections::HashMap;
 
+use bounded_vec::{BoundedVec, BoundedVecOutOfBounds};
 use ergotree_ir::chain::ergo_box::box_value::BoxValue;
 use ergotree_ir::chain::ergo_box::box_value::BoxValueError;
 use ergotree_ir::chain::ergo_box::BoxId;
+use ergotree_ir::chain::ergo_box::BoxTokens;
 use ergotree_ir::chain::ergo_box::ErgoBox;
 use ergotree_ir::chain::ergo_box::ErgoBoxCandidate;
 use ergotree_ir::chain::token::Token;
 use ergotree_ir::chain::token::TokenAmount;
+use ergotree_ir::chain::token::TokenAmountError;
 use ergotree_ir::chain::token::TokenId;
 pub use simple::*;
 
 use thiserror::Error;
 
+/// Bounded vec with minimum 1 element and max u16::MAX elements
+pub type SelectedBoxes<T> = BoundedVec<T, 1, { u16::MAX as usize }>;
+
 /// Selected boxes (by [`BoxSelector`])
 #[derive(PartialEq, Eq, Debug, Clone)]
 pub struct BoxSelection<T: ErgoBoxAssets> {
     /// Selected boxes to spend as transaction inputs
-    pub boxes: Vec<T>,
+    pub boxes: SelectedBoxes<T>,
     /// box assets with returning change amounts (to be put in tx outputs)
     pub change_boxes: Vec<ErgoBoxAssetsData>,
 }
@@ -52,6 +59,14 @@ pub enum BoxSelectorError {
     /// BoxValue out of bounds
     #[error("BoxValue out of bounds")]
     BoxValueError(BoxValueError),
+
+    /// Token amount err
+    #[error("TokenAmountError: {0:?}")]
+    TokenAmountError(#[from] TokenAmountError),
+
+    /// Boxes out of bounds
+    #[error("Boxes is out of bounds")]
+    OutOfBounds(#[from] BoundedVecOutOfBounds),
 }
 
 impl From<BoxValueError> for BoxSelectorError {
@@ -65,7 +80,7 @@ pub trait ErgoBoxAssets {
     /// Box value
     fn value(&self) -> BoxValue;
     /// Tokens (ids and amounts)
-    fn tokens(&self) -> Vec<Token>;
+    fn tokens(&self) -> Option<BoxTokens>;
 }
 
 /// Simple struct to hold ErgoBoxAssets values
@@ -74,7 +89,7 @@ pub struct ErgoBoxAssetsData {
     /// Box value
     pub value: BoxValue,
     /// Tokens
-    pub tokens: Vec<Token>,
+    pub tokens: Option<BoxTokens>,
 }
 
 impl ErgoBoxAssets for ErgoBoxAssetsData {
@@ -82,7 +97,7 @@ impl ErgoBoxAssets for ErgoBoxAssetsData {
         self.value
     }
 
-    fn tokens(&self) -> Vec<Token> {
+    fn tokens(&self) -> Option<BoxTokens> {
         self.tokens.clone()
     }
 }
@@ -92,7 +107,7 @@ impl ErgoBoxAssets for ErgoBoxCandidate {
         self.value
     }
 
-    fn tokens(&self) -> Vec<Token> {
+    fn tokens(&self) -> Option<BoxTokens> {
         self.tokens.clone()
     }
 }
@@ -102,7 +117,7 @@ impl ErgoBoxAssets for ErgoBox {
         self.value
     }
 
-    fn tokens(&self) -> Vec<Token> {
+    fn tokens(&self) -> Option<BoxTokens> {
         self.tokens.clone()
     }
 }
@@ -125,36 +140,47 @@ pub fn sum_value<T: ErgoBoxAssets>(bs: &[T]) -> u64 {
 }
 
 /// Returns the total token amounts (all tokens combined)
-pub fn sum_tokens(ts: &[Token]) -> HashMap<TokenId, TokenAmount> {
+pub fn sum_tokens(ts: Option<&[Token]>) -> Result<HashMap<TokenId, TokenAmount>, TokenAmountError> {
     let mut res: HashMap<TokenId, TokenAmount> = HashMap::new();
-    ts.iter().for_each(|t| {
-        res.entry(t.token_id.clone())
-            .and_modify(|amt| *amt = amt.checked_add(&t.amount).unwrap())
-            .or_insert(t.amount);
-    });
-    res
+    ts.into_iter().flatten().try_for_each(|t| {
+        if let Some(amt) = res.get_mut(&t.token_id) {
+            *amt = amt.checked_add(&t.amount)?;
+        } else {
+            res.insert(t.token_id.clone(), t.amount);
+        }
+        Ok(())
+    })?;
+    Ok(res)
 }
 
 /// Returns the total token amounts (all tokens combined) of the given boxes
-pub fn sum_tokens_from_boxes<T: ErgoBoxAssets>(bs: &[T]) -> HashMap<TokenId, TokenAmount> {
+pub fn sum_tokens_from_boxes<T: ErgoBoxAssets>(
+    bs: &[T],
+) -> Result<HashMap<TokenId, TokenAmount>, TokenAmountError> {
     let mut res: HashMap<TokenId, TokenAmount> = HashMap::new();
-    bs.iter().for_each(|b| {
-        b.tokens().iter().for_each(|t| {
-            res.entry(t.token_id.clone())
-                .and_modify(|amt| *amt = amt.checked_add(&t.amount).unwrap())
-                .or_insert(t.amount);
-        });
-    });
-    res
+    bs.iter().try_for_each(|b| {
+        b.tokens().into_iter().flatten().try_for_each(|t| {
+            if let Some(amt) = res.get_mut(&t.token_id) {
+                *amt = amt.checked_add(&t.amount)?;
+            } else {
+                res.insert(t.token_id.clone(), t.amount);
+            }
+
+            Ok(())
+        })
+    })?;
+    Ok(res)
 }
 
 #[cfg(test)]
+#[allow(clippy::unwrap_used, clippy::panic)]
 mod tests {
 
     use ergotree_ir::chain::ergo_box::box_value::arbitrary::ArbBoxValueRange;
     use ergotree_ir::chain::ergo_box::box_value::BoxValue;
+    use ergotree_ir::chain::ergo_box::BoxTokens;
     use ergotree_ir::chain::token::Token;
-    use proptest::{arbitrary::Arbitrary, collection::vec, prelude::*};
+    use proptest::{arbitrary::Arbitrary, collection::vec, option::of, prelude::*};
     use sigma_test_util::force_any_val;
 
     use crate::wallet::box_selector::sum_tokens;
@@ -166,8 +192,11 @@ mod tests {
         type Parameters = ArbBoxValueRange;
 
         fn arbitrary_with(args: Self::Parameters) -> Self::Strategy {
-            (any_with::<BoxValue>(args), vec(any::<Token>(), 0..3))
-                .prop_map(|(value, tokens)| Self { value, tokens })
+            (any_with::<BoxValue>(args), of(vec(any::<Token>(), 0..3)))
+                .prop_map(|(value, tokens)| Self {
+                    value,
+                    tokens: tokens.map(BoxTokens::from_vec).and_then(Result::ok),
+                })
                 .boxed()
         }
 
@@ -183,11 +212,12 @@ mod tests {
         let token = force_any_val::<Token>();
         let b = ErgoBoxAssetsData {
             value: BoxValue::SAFE_USER_MIN,
-            tokens: vec![token.clone(), token.clone()],
+            tokens: BoxTokens::from_vec(vec![token.clone(), token.clone()]).ok(),
         };
         assert_eq!(
             u64::from(
                 *sum_tokens_from_boxes(vec![b.clone(), b].as_slice())
+                    .unwrap()
                     .get(&token.token_id)
                     .unwrap()
             ),
@@ -199,7 +229,7 @@ mod tests {
 
         #[test]
         fn sum_tokens_eq(b in any::<ErgoBoxAssetsData>()) {
-            prop_assert_eq!(sum_tokens(b.tokens.as_slice()), sum_tokens_from_boxes(vec![b].as_slice()))
+            prop_assert_eq!(sum_tokens(b.tokens.as_ref().map(BoxTokens::as_ref)), sum_tokens_from_boxes(vec![b].as_slice()))
         }
     }
 }
