@@ -1,6 +1,9 @@
 //! Builder for an UnsignedTransaction
 
+use ergotree_interpreter::sigma_protocol::prover::ContextExtension;
 use ergotree_ir::chain::ergo_box::box_value::BoxValueError;
+use ergotree_ir::chain::token::TokenAmountError;
+use std::collections::HashMap;
 use std::collections::HashSet;
 use std::convert::TryInto;
 
@@ -40,6 +43,7 @@ pub struct TxBuilder<S: ErgoBoxAssets> {
     fee_amount: BoxValue,
     change_address: Address,
     min_change_value: BoxValue,
+    context_extensions: HashMap<BoxId, ContextExtension>,
 }
 
 impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
@@ -67,6 +71,7 @@ impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
             fee_amount,
             change_address,
             min_change_value,
+            context_extensions: HashMap::new(),
         }
     }
 
@@ -110,6 +115,11 @@ impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
         self.data_inputs = data_inputs;
     }
 
+    /// Set context extension for a given input
+    pub fn set_context_extension(&mut self, box_id: BoxId, context_extension: ContextExtension) {
+        self.context_extensions.insert(box_id, context_extension);
+    }
+
     /// Estimated serialized transaction size in bytes after signing (assuming P2PK box spending)
     pub fn estimate_tx_size_bytes(&self) -> Result<usize, TxBuilderError> {
         let tx = self.build_tx()?;
@@ -169,7 +179,7 @@ impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
                     change_address_ergo_tree.clone(),
                     self.current_height,
                 );
-                for token in &b.tokens() {
+                for token in b.tokens().into_iter().flatten() {
                     candidate.add_token(token.clone());
                 }
                 candidate.build()
@@ -199,9 +209,9 @@ impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
             ));
         }
         // check that inputs have enough tokens
-        let input_tokens = sum_tokens_from_boxes(self.box_selection.boxes.as_slice());
-        let output_tokens = sum_tokens_from_boxes(output_candidates.as_slice());
-        let first_input_box_id: TokenId = self.box_selection.boxes.first().unwrap().box_id().into();
+        let input_tokens = sum_tokens_from_boxes(self.box_selection.boxes.as_slice())?;
+        let output_tokens = sum_tokens_from_boxes(output_candidates.as_slice())?;
+        let first_input_box_id: TokenId = self.box_selection.boxes.first().box_id().into();
         let output_tokens_len = output_tokens.len();
         let output_tokens_without_minted: Vec<Token> = output_tokens
             .into_iter()
@@ -222,14 +232,16 @@ impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
                 }
             })?;
 
+        let unsigned_inputs = self.box_selection.boxes.clone().mapped(|b| {
+            let ctx_ext = self
+                .context_extensions
+                .get(&b.box_id())
+                .cloned()
+                .unwrap_or_else(ContextExtension::empty);
+            UnsignedInput::new(b.box_id(), ctx_ext)
+        });
         Ok(UnsignedTransaction::new(
-            self.box_selection
-                .boxes
-                .clone()
-                .into_iter()
-                .map(UnsignedInput::from)
-                .collect::<Vec<UnsignedInput>>()
-                .try_into()?,
+            unsigned_inputs,
             self.data_inputs.clone().try_into().ok(),
             output_candidates.try_into()?,
         )?)
@@ -242,12 +254,13 @@ impl<S: ErgoBoxAssets + ErgoBoxId + Clone> TxBuilder<S> {
 }
 
 /// Suggested transaction fee (1100000 nanoERGs, semi-default value used across wallets and dApps as of Oct 2020)
-#[allow(non_snake_case)]
+#[allow(non_snake_case, clippy::unwrap_used)]
 pub fn SUGGESTED_TX_FEE() -> BoxValue {
     BoxValue::new(1100000u64).unwrap()
 }
 
 /// Create a box with miner's contract and a given value
+#[allow(clippy::unwrap_used)]
 pub fn new_miner_fee_box(
     fee_amount: BoxValue,
     creation_height: u32,
@@ -256,7 +269,9 @@ pub fn new_miner_fee_box(
     let miner_fee_address = address_encoder
         .parse_address_from_str(MINERS_FEE_MAINNET_ADDRESS)
         .unwrap();
-    let ergo_tree = miner_fee_address.script().unwrap();
+    let ergo_tree = miner_fee_address
+        .script()
+        .map_err(ErgoBoxCandidateBuilderError::ParsingError)?;
     ErgoBoxCandidateBuilder::new(fee_amount, ergo_tree, creation_height).build()
 }
 
@@ -290,9 +305,16 @@ pub enum TxBuilderError {
     /// Invalid Tx input count
     #[error("Invalid tx inputs count: {0}")]
     InvalidInputsCount(#[from] BoundedVecOutOfBounds),
+    /// Input box was unable to be retrieved
+    #[error("Empty input box")]
+    EmptyInputBoxSelection,
+    /// Token amount err
+    #[error("TokenAmountError: {0:?}")]
+    TokenAmountError(#[from] TokenAmountError),
 }
 
 #[cfg(test)]
+#[allow(clippy::unwrap_used, clippy::panic)]
 mod tests {
 
     use std::convert::TryInto;
@@ -312,28 +334,11 @@ mod tests {
 
     use super::*;
 
-    #[test]
-    fn test_empty_inputs() {
-        let box_selection: BoxSelection<ErgoBox> = BoxSelection {
-            boxes: vec![],
-            change_boxes: vec![],
-        };
-        let r = TxBuilder::new(
-            box_selection,
-            vec![force_any_val::<ErgoBoxCandidate>()],
-            1,
-            force_any_val::<BoxValue>(),
-            force_any_val::<Address>(),
-            BoxValue::SAFE_USER_MIN,
-        );
-        assert!(r.build().is_err());
-    }
-
     #[test]
     fn test_duplicate_inputs() {
         let input_box = force_any_val::<ErgoBox>();
         let box_selection: BoxSelection<ErgoBox> = BoxSelection {
-            boxes: vec![input_box.clone(), input_box],
+            boxes: vec![input_box.clone(), input_box].try_into().unwrap(),
             change_boxes: vec![],
         };
         let r = TxBuilder::new(
@@ -373,7 +378,7 @@ mod tests {
         let input_box = ErgoBox::new(
             10000000i64.try_into().unwrap(),
             force_any_val::<ErgoTree>(),
-            vec![token_pair.clone()],
+            vec![token_pair.clone()].try_into().ok(),
             NonMandatoryRegisters::empty(),
             1,
             force_any_val::<TxId>(),
@@ -405,7 +410,7 @@ mod tests {
         );
         let tx = tx_builder.build().unwrap();
         assert!(
-            tx.output_candidates.get(0).unwrap().tokens().is_empty(),
+            tx.output_candidates.get(0).unwrap().tokens().is_none(),
             "expected empty tokens in the first output box"
         );
     }
@@ -415,7 +420,7 @@ mod tests {
         let input_box = ErgoBox::new(
             100000000i64.try_into().unwrap(),
             force_any_val::<ErgoTree>(),
-            vec![],
+            None,
             NonMandatoryRegisters::empty(),
             1,
             force_any_val::<TxId>(),
@@ -456,8 +461,8 @@ mod tests {
                 .get(0)
                 .unwrap()
                 .tokens()
-                .first()
                 .unwrap()
+                .first()
                 .token_id,
             token_pair.token_id,
             "expected minted token in the first output box"
@@ -515,7 +520,7 @@ mod tests {
         let inputs: Vec<ErgoBox> = vec![input_box];
         let tx_fee = BoxValue::SAFE_USER_MIN;
         let box_selection = BoxSelection {
-            boxes: inputs,
+            boxes: inputs.try_into().unwrap(),
             change_boxes: vec![],
         };
         let outputs = vec![out_box];
@@ -538,7 +543,7 @@ mod tests {
         let input = ErgoBox::new(
             10000000i64.try_into().unwrap(),
             force_any_val::<ErgoTree>(),
-            vec![],
+            None,
             NonMandatoryRegisters::empty(),
             1,
             force_any_val::<TxId>(),
@@ -553,7 +558,7 @@ mod tests {
         let outputs = vec![out_box];
         let tx_builder = TxBuilder::new(
             BoxSelection {
-                boxes: vec![input],
+                boxes: vec![input].try_into().unwrap(),
                 change_boxes: vec![],
             },
             outputs,
@@ -574,18 +579,21 @@ mod tests {
                          outputs in vec(any_with::<ErgoBoxCandidate>((BoxValue::MIN_RAW * 1000 ..BoxValue::MIN_RAW * 2000).into()), 1..2),
                          change_address in any::<Address>(),
                          miners_fee in any_with::<BoxValue>((BoxValue::MIN_RAW * 100..BoxValue::MIN_RAW * 200).into()),
-                         data_inputs in vec(any::<DataInput>(), 0..2)) {
-            prop_assume!(sum_tokens_from_boxes(outputs.as_slice()).is_empty());
+                         data_inputs in vec(any::<DataInput>(), 0..2),
+                         ctx_ext in any::<ContextExtension>()) {
+            prop_assume!(sum_tokens_from_boxes(outputs.as_slice()).unwrap().is_empty());
             let min_change_value = BoxValue::SAFE_USER_MIN;
             let all_outputs = checked_sum(outputs.iter().map(|b| b.value)).unwrap()
-                                                                             .checked_add(&miners_fee)
-                                                                             .unwrap();
+                .checked_add(&miners_fee)
+                .unwrap();
             let all_inputs = checked_sum(inputs.iter().map(|b| b.value)).unwrap();
             prop_assume!(all_outputs < all_inputs);
-            let total_output_value: BoxValue = checked_sum(outputs.iter().map(|b| b.value)).unwrap()
-                                                                                                      .checked_add(&miners_fee).unwrap();
+            let total_output_value: BoxValue = checked_sum(outputs.iter().map(|b| b.value))
+                .unwrap()
+                .checked_add(&miners_fee).unwrap();
+            let selection = SimpleBoxSelector::new().select(inputs.clone(), total_output_value, &[]).unwrap();
             let mut tx_builder = TxBuilder::new(
-                SimpleBoxSelector::new().select(inputs.clone(), total_output_value, &[]).unwrap(),
+                selection.clone(),
                 outputs.clone(),
                 1,
                 miners_fee,
@@ -593,6 +601,7 @@ mod tests {
                 min_change_value,
             );
             tx_builder.set_data_inputs(data_inputs.clone());
+            tx_builder.set_context_extension(selection.boxes.first().box_id(), ctx_ext.clone());
             let tx = tx_builder.build().unwrap();
             prop_assert!(outputs.into_iter().all(|i| tx.output_candidates.iter().any(|o| *o == i)),
                          "tx.output_candidates is missing some outputs");
@@ -608,6 +617,7 @@ mod tests {
                 b.value == miners_fee
             }), "box with miner's fee {:?} is not found in outputs: {:?}", miners_fee, tx.output_candidates);
             prop_assert_eq!(tx.data_inputs.map(|i| i.as_vec().clone()).unwrap_or_default(), data_inputs, "unexpected data inputs");
+            prop_assert_eq!(&tx.inputs.first().extension, &ctx_ext);
         }
     }
 }
